import logging
import re
import math
from pathlib import Path
import pygments

from gi.repository import Gtk
from gi.repository import Gdk
from gi.repository.Pango import FontDescription
from gi.repository.Pango import Style
from gi.repository.Pango import SCALE

from gajim.common import app

from gajim.plugins.plugins_i18n import _
from gajim.plugins.helpers import get_builder

from syntax_highlight.gtkformatter import GTKFormatter
from syntax_highlight.types import LineBreakOptions
from syntax_highlight.types import CodeMarkerOptions
from syntax_highlight.types import PLUGIN_INTERNAL_NONE_LEXER_ID

log = logging.getLogger('gajim.p.syntax_highlight')

PLUGIN_INTERNAL_NONE_LEXER = ('None (monospace only)',
                              PLUGIN_INTERNAL_NONE_LEXER_ID)


class SyntaxHighlighterPluginConfig(Gtk.ApplicationWindow):
    def __init__(self, plugin, transient):
        Gtk.ApplicationWindow.__init__(self)
        self.set_application(app.app)
        self.set_show_menubar(False)
        self.set_title(_('Syntax Highlighter Configuration'))
        self.set_transient_for(transient)
        self.set_default_size(400, 500)
        self.set_type_hint(Gdk.WindowTypeHint.DIALOG)
        self.set_modal(True)
        self.set_destroy_with_parent(True)

        ui_path = Path(__file__).parent
        self._ui = get_builder(ui_path.resolve() / 'config_dialog.ui')
        self.add(self._ui.main_box)
        self.show_all()

        self._ui.preview_textview.get_buffer().connect(
            'insert-text', self._on_preview_text_inserted)
        self._ui.connect_signals(self)

        self._lexer_liststore = Gtk.ListStore(str)
        self._ui.default_lexer_combobox.set_model(self._lexer_liststore)

        self._style_liststore = Gtk.ListStore(str)
        self._ui.style_combobox.set_model(self._style_liststore)

        self._plugin = plugin
        self._lexers = plugin.highlighter_config.get_lexer_list()
        self._styles = plugin.highlighter_config.get_styles_list()

        self._provider = None
        self._add_css_provider()

        self._initialize()

    def _initialize(self):
        default_lexer = self._plugin.highlighter_config.get_default_lexer_name()

        for i, lexer in enumerate(self._lexers):
            self._lexer_liststore.append([lexer[0]])
            if lexer[1] == default_lexer:
                self._ui.default_lexer_combobox.set_active(i)

        for i, style in enumerate(self._styles):
            self._style_liststore.append([style])
            if style == self._plugin.config['style']:
                self._ui.style_combobox.set_active(i)

        self._ui.line_break_combobox.set_active(
            self._plugin.config['line_break'].value)
        self._ui.code_marker_combobox.set_active(
            self._plugin.config['code_marker'])
        self._ui.font_button.set_font(self._plugin.config['font'])

        bg_override_enabled = self._plugin.config['bgcolor_override']
        self._ui.bg_color_checkbutton.set_active(bg_override_enabled)
        self._ui.bg_color_colorbutton.set_sensitive(bg_override_enabled)
        color = Gdk.RGBA()
        if color.parse(self._plugin.config['bgcolor']):
            self._ui.bg_color_colorbutton.set_rgba(color)
        self._update_preview()

    def _lexer_changed(self, widget):
        self._plugin.highlighter_config.set_default_lexer(
            self._lexers[widget.get_active()][1])
        self._update_preview()

    def _line_break_changed(self, widget):
        self._plugin.config['line_break'] = LineBreakOptions(
            widget.get_active())
        self._update_preview()

    def _code_marker_changed(self, widget):
        self._plugin.config['code_marker'] = CodeMarkerOptions(
            widget.get_active())

    def _bg_color_enabled(self, widget):
        override_color = widget.get_active()
        self._plugin.config['bgcolor_override'] = override_color
        self._ui.bg_color_colorbutton.set_sensitive(override_color)
        self._update_preview()

    def _bg_color_changed(self, widget):
        color = widget.get_rgba()
        self._plugin.config['bgcolor'] = color.to_string()
        self._update_preview()

    def _style_changed(self, widget):
        style = self._styles[widget.get_active()]
        if style is not None and style != '':
            self._plugin.config['style'] = style
        self._update_preview()

    def _font_changed(self, widget):
        font = widget.get_font()
        if font is not None and font != '':
            self._plugin.config['font'] = font
        self._update_preview()

    def _update_preview(self):
        self._format_preview_text()

    def _on_preview_text_inserted(self, _buf, _iterator, text, length, *_args):
        if (length == 1 and re.match(r'\s', text)) or length > 1:
            self._format_preview_text()

    def _add_css_provider(self):
        self._context = self._ui.preview_textview.get_style_context()
        self._provider = Gtk.CssProvider()
        self._context.add_provider(
            self._provider, Gtk.STYLE_PROVIDER_PRIORITY_USER)
        self._context.add_class('syntax-preview')

    def _format_preview_text(self):
        buf = self._ui.preview_textview.get_buffer()
        start_iter = buf.get_start_iter()
        start_mark = buf.create_mark(None, start_iter, True)
        buf.remove_all_tags(start_iter, buf.get_end_iter())

        formatter = GTKFormatter(
            style=self._plugin.config['style'], start_mark=start_mark)

        code = start_iter.get_text(buf.get_end_iter())
        lexer = self._plugin.highlighter_config.get_default_lexer()
        if lexer != PLUGIN_INTERNAL_NONE_LEXER_ID:
            tokens = pygments.lex(code, lexer)
            pygments.format(tokens, formatter, buf)

        buf.delete_mark(start_mark)
        css = self._get_css()
        self._provider.load_from_data(bytes(css.encode()))

    def _get_css(self):
        # Build CSS from Pango.FontDescription
        description = FontDescription.from_string(self._plugin.config['font'])
        size = description.get_size() / SCALE
        style = self._get_string_from_pango_style(description.get_style())
        weight = self._pango_to_css_weight(int(description.get_weight()))
        family = description.get_family()
        font = '%spt %s' % (size, family)

        if self._plugin.config['bgcolor_override']:
            color = self._plugin.config['bgcolor']
        else:
            color = '@theme_base_color'

        css = '''
        .syntax-preview {
            font: %s;
            font-weight: %s;
            font-style: %s;
        }
        .syntax-preview > text {
            background-color: %s;
        }
        ''' % (font, weight, style, color)
        return css

    @staticmethod
    def _pango_to_css_weight(number):
        # Pango allows for weight values between 100 and 1000
        # CSS allows only full hundred numbers like 100, 200 ..
        number = int(number)
        if number < 100:
            return 100
        if number > 900:
            return 900
        return int(math.ceil(number / 100.0)) * 100

    @staticmethod
    def _get_string_from_pango_style(style: Style) -> str:
        if style == Style.NORMAL:
            return 'normal'
        if style == Style.ITALIC:
            return 'italic'
        # Style.OBLIQUE:
        return 'oblique'



from enum import Enum
from enum import IntEnum
from enum import unique

PLUGIN_INTERNAL_NONE_LEXER_ID = '_syntax_highlight_internal_none_type'


class MatchType(Enum):
    INLINE = 0
    MULTILINE = 1
    TEXT = 2


@unique
class LineBreakOptions(IntEnum):
    NEVER = 0
    ALWAYS = 1
    MULTILINE = 2


@unique
class CodeMarkerOptions(IntEnum):
    AS_COMMENT = 0
    HIDE = 1



import logging

from gi.repository import Gtk
from gi.repository import Pango

from pygments.formatter import Formatter

log = logging.getLogger('gajim.p.syntax_highlight')


class GTKFormatter(Formatter):
    name = 'GTK Formatter'
    aliases = ['textbuffer', 'gtk']
    filenames = []

    def __init__(self, **options):
        super(GTKFormatter, self).__init__(**options)
        # Formatter.__init__(self, **options)
        self.tags = {}
        self.mark = options.get('start_mark', None)

    @staticmethod
    def create_tag_for_token(ttype, highlighting_style):
        style = highlighting_style.style_for_token(ttype)
        tag = Gtk.TextTag.new()
        if 'bgcolor' in style and not style['bgcolor'] is None:
            tag.set_property('background', '#%s' % style['bgcolor'])
        if 'bold' in style and style['bold']:
            tag.set_property('weight', Pango.Weight.BOLD)
        if 'border' in style and not style['border'] is None:
            pass
        if 'color' in style and not style['color'] is None:
            tag.set_property('foreground', '#%s' % style['color'])
        if 'italic' in style and style['italic']:
            tag.set_property('style', Pango.Style.ITALIC)
        if 'mono' in style and not style['mono'] is None:
            tag.set_property('family', 'Monospace')
            tag.set_property('family-set', True)
        if 'roman' in style and not style['roman'] is None:
            pass
        if 'sans' in style and not style['sans'] is None:
            tag.set_property('family', 'Sans')
            tag.set_property('family-set', True)
        if 'underline' in style and style['underline']:
            tag.set_property('underline', 'single')
        return tag

    def get_tag(self, ttype, buf):
        """
        Creates, stores and returs a tag for a given token type.

        This method ensures that a tag style is created only once.
        Furthermore, the tag will be added to the given Gtk.TextBuffer's
        tag table.
        """
        tag = None
        if ttype in self.tags:
            tag = self.tags[ttype]
        else:
            tag = GTKFormatter.create_tag_for_token(ttype, self.style)
            self.tags[ttype] = tag
            buf.get_tag_table().add(tag)
        return tag

    def set_insert_pos_mark(self, mark):
        self.mark = mark

    def format(self, tokensource, outfile):
        if not isinstance(outfile, Gtk.TextBuffer) or outfile is None:
            log.warning('Did not get a buffer to format...')
            return
        buf = outfile

        end_iter = buf.get_end_iter()

        start_mark = self.mark
        start_iter = buf.get_iter_at_mark(start_mark) if \
            start_mark is not None else end_iter

        last_ttype = None
        last_start = start_iter
        last_end = buf.get_end_iter()
        last_fixed_start = last_start

        reset = True

        for ttype, value in tokensource:
            search = None
            if last_ttype is not None and ttype != last_ttype:
                tag = self.get_tag(last_ttype, buf)

                buf.apply_tag(tag, last_fixed_start, last_end)

                search = last_end.forward_search(
                    value, Gtk.TextSearchFlags.TEXT_ONLY, end_iter)
                reset = True
            else:
                # In case last_ttype is None, i.e. first loop walkthrough:
                # last_start to end_iter is the full code block.
                search = last_start.forward_search(
                    value, Gtk.TextSearchFlags.TEXT_ONLY, end_iter)

            # Prepare for next iteration
            last_ttype = ttype
            if search is not None:
                (last_start, last_end) = search

                # If we've found the end of a sequence of similar type tokens
                # or if we are in the first loop iteration, set the fixed point
                if reset:
                    last_fixed_start = last_start
                    reset = False
            else:
                # Hm... Nothing found, but tags left? Seems there's nothing we
                # can do now.
                break



import logging
from functools import partial

from gajim.plugins import GajimPlugin
from gajim.plugins.plugins_i18n import _

from syntax_highlight.types import LineBreakOptions
from syntax_highlight.types import CodeMarkerOptions
from syntax_highlight.types import PLUGIN_INTERNAL_NONE_LEXER_ID

log = logging.getLogger('gajim.p.syntax_highlight')

HAS_PYGMENTS = False
try:
    from syntax_highlight.chat_syntax_highlighter import ChatSyntaxHighlighter
    from syntax_highlight.config_dialog import SyntaxHighlighterPluginConfig
    from syntax_highlight.highlighter_config import HighlighterConfig
    HAS_PYGMENTS = True
except Exception as exception:
    log.error('Could not load pygments: %s', exception)


class SyntaxHighlighterPlugin(GajimPlugin):
    def init(self):
        self.description = _(
            'Source code syntax highlighting in the chat window.\n\n'
            'Markdown-style syntax is supported, i.e. text inbetween '
            '`single backticks` is rendered as inline code.\n'
            '```language\n'
            'selection is possible in multi-line code snippets inbetween '
            'triple-backticks\n'
            'Note the newlines in this case…\n'
            '```\n\n'
            'Changed settings will take effect after re-opening the message '
            'tab/window.')

        self.config_default_values = {
            'default_lexer': (PLUGIN_INTERNAL_NONE_LEXER_ID, ''),
            'line_break': (LineBreakOptions.MULTILINE, ''),
            'style': ('default', ''),
            'font': ('Monospace 10', ''),
            'bgcolor': ('rgb(200, 200, 200)', ''),
            'bgcolor_override': (True, ''),
            'code_marker': (CodeMarkerOptions.AS_COMMENT, ''),
        }

        self.gui_extension_points = {
            'chat_control_base': (
                self._connect_chat_control,
                self._disconnect_chat_control),
            'print_real_text': (self._on_print_real_text, None)
        }

        if not HAS_PYGMENTS:
            self.activatable = False
            self.available_text = _('You are missing python-pygments.')
            self.config_dialog = None

        self._migrate_settings()
        self._highlighters = {}
        self.config_dialog = partial(SyntaxHighlighterPluginConfig, self)
        self.highlighter_config = HighlighterConfig(self.config)

    def _migrate_settings(self):
        line_break = self.config['line_break']
        if isinstance(line_break, int):
            self.config['line_break'] = LineBreakOptions(line_break)

    def _connect_chat_control(self, chat_control):
        highlighter = ChatSyntaxHighlighter(
            self.config, self.highlighter_config, chat_control)
        self._highlighters[chat_control.control_id] = highlighter

    def _disconnect_chat_control(self, chat_control):
        highlighter = self._highlighters.get(chat_control.control_id)
        if highlighter is not None:
            del highlighter
            self._highlighters.pop(chat_control.control_id, None)

    def _on_print_real_text(self, text_view, real_text, other_tags, graphics,
                            iterator, additional):
        for highlighter in self._highlighters.values():
            if highlighter.textview != text_view:
                continue
            highlighter.process_text(
                real_text, other_tags, graphics, iterator, additional)
            return

    def update_highlighters(self):
        for highlighter in self._highlighters.values():
            highlighter.update_config(self.config)



import logging

from pygments.lexers import get_lexer_by_name
from pygments.lexers import get_all_lexers
from pygments.styles import get_all_styles
from pygments.util import ClassNotFound

from syntax_highlight.types import PLUGIN_INTERNAL_NONE_LEXER_ID

log = logging.getLogger('gajim.p.syntax_highlight')
PLUGIN_INTERNAL_NONE_LEXER = ('None (monospace only)',
                              PLUGIN_INTERNAL_NONE_LEXER_ID)


class HighlighterConfig:
    def __init__(self, plugin_config):
        self._plugin_config = plugin_config

        self._lexer_list = self._create_lexer_list()
        self._style_list = []
        for style in get_all_styles():
            self._style_list.append(style)
        self._style_list.sort()

        self._default_lexer = None
        self.set_default_lexer(self._plugin_config['default_lexer'])

    @staticmethod
    def _create_lexer_list():
        # The list we create here contains the plain text name and the lexer's
        # id string
        lexers = []

        # Iteration over get_all_lexers() seems to be broken somehow
        # Workaround
        all_lexers = get_all_lexers()
        for lexer in all_lexers:
            # We don't want to add lexers that we cant identify by name later
            if lexer[1] is not None and lexer[1]:
                lexers.append((lexer[0], lexer[1][0]))
        lexers.sort()

        # Insert our internal 'none' type at top of the list
        lexers.insert(0, PLUGIN_INTERNAL_NONE_LEXER)
        return lexers

    @staticmethod
    def get_lexer_by_name(name):
        lexer = None
        try:
            lexer = get_lexer_by_name(name)
        except ClassNotFound:
            pass
        return lexer

    def get_lexer_with_fallback(self, language):
        lexer = self.get_lexer_by_name(language)
        if lexer is None:
            log.info('Falling back to default lexer for %s.',
                     self.get_default_lexer_name())
            lexer = self._default_lexer[1]
        return lexer

    def set_default_lexer(self, name):
        if name != PLUGIN_INTERNAL_NONE_LEXER_ID:
            lexer = get_lexer_by_name(name)

            if lexer is None and self._default_lexer is None:
                log.error('Failed to get default lexer by name.'
                          'Falling back to simply using the first lexer '
                          'in the list.')
                lexer = self._lexer_list[0]
                name = lexer[0]
                self._default_lexer = (name, lexer)
            if lexer is None and self._default_lexer is not None:
                log.info('Failed to get default lexer by name, keeping '
                         'previous setting (lexer = %s).',
                         self._default_lexer[0])
                name = self._default_lexer[0]
            else:
                self._default_lexer = (name, lexer)
        else:
            self._default_lexer = PLUGIN_INTERNAL_NONE_LEXER

        self._plugin_config['default_lexer'] = name

    def get_default_lexer(self):
        return self._default_lexer[1]

    def get_default_lexer_name(self):
        return self._default_lexer[0]

    def get_lexer_list(self):
        return self._lexer_list

    def get_styles_list(self):
        return self._style_list



import logging
import re
from packaging.version import Version as V
import pygments
from pygments.token import Comment
from pygments.styles import get_style_by_name

from gi.repository import Gtk
from gi.repository import Gdk

import gajim

from gajim.plugins.plugins_i18n import _

from syntax_highlight.gtkformatter import GTKFormatter
from syntax_highlight.types import MatchType
from syntax_highlight.types import LineBreakOptions
from syntax_highlight.types import CodeMarkerOptions
from syntax_highlight.types import PLUGIN_INTERNAL_NONE_LEXER_ID

log = logging.getLogger('gajim.p.syntax_highlight')


class ChatSyntaxHighlighter:
    def __init__(self, plugin_config, highlighter_config, chat_control):
        self.textview = chat_control.conv_textview
        self._plugin_config = plugin_config
        self._highlighter_config = highlighter_config
        self._chat_control = chat_control
        self._chat_control.msg_textview.connect(
            'populate-popup', self._on_msg_textview_populate_popup)

    def update_config(self, plugin_config):
        self._plugin_config = plugin_config

    def _on_msg_textview_populate_popup(self, _textview, menu):
        item = Gtk.MenuItem.new_with_mnemonic(_('_Paste as Code'))
        menu.append(item)
        id_ = item.connect('activate', self._paste_as_code)
        self._chat_control.handlers[id_] = item

        item = Gtk.MenuItem.new_with_mnemonic(_('Paste as Code _Block'))
        menu.append(item)
        id_ = item.connect('activate', self._paste_as_code_block)
        self._chat_control.handlers[id_] = item

        menu.show_all()

    @staticmethod
    def _get_clipboard_text():
        clipboard = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)
        return clipboard.wait_for_text()

    def _insert_paste(self, text):
        if V(gajim.__version__) <= V('1.3.1'):
            self._chat_control.msg_textview.remove_placeholder()
        message_buffer = self._chat_control.msg_textview.get_buffer()
        message_buffer.insert_at_cursor(text)

    def _paste_as_code(self, _item):
        text = self._get_clipboard_text()
        self._insert_paste(f'`{text}`')

    def _paste_as_code_block(self, _item):
        text = self._get_clipboard_text()
        self._insert_paste(f'```\n{text}\n```')

    @staticmethod
    def _hide_code_markup(buf, start, end):
        tag = buf.get_tag_table().lookup('hide_code_markup')
        if tag is None:
            tag = Gtk.TextTag.new('hide_code_markup')
            tag.set_property('invisible', True)
            buf.get_tag_table().add(tag)

        buf.apply_tag_by_name('hide_code_markup', start, end)

    def _check_line_break(self, is_multiline):
        line_break = self._plugin_config['line_break'].value
        return (line_break == LineBreakOptions.ALWAYS) \
            or (is_multiline and line_break == LineBreakOptions.MULTILINE)

    def _format_code(self, buf, s_tag, s_code, e_tag, e_code, language):
        style = self._plugin_config['style']
        if self._plugin_config['code_marker'] == CodeMarkerOptions.HIDE:
            self._hide_code_markup(buf, s_tag, s_code)
            self._hide_code_markup(buf, e_code, e_tag)
        else:
            comment_tag = GTKFormatter.create_tag_for_token(
                Comment, get_style_by_name(style))
            buf.get_tag_table().add(comment_tag)
            buf.apply_tag(comment_tag, s_tag, s_code)
            buf.apply_tag(comment_tag, e_tag, e_code)

        code = s_code.get_text(e_code)
        log.debug('full text to encode: %s.', code)

        start_mark = buf.create_mark(None, s_code, False)

        lexer = None

        if language is None:
            lexer = self._highlighter_config.get_default_lexer()
            log.info('No Language specified. '
                     'Falling back to default lexer: %s.',
                     self._highlighter_config.get_default_lexer_name())
        else:
            log.debug('Using lexer for %s.', str(language))
            lexer = self._highlighter_config.get_lexer_with_fallback(language)

        if lexer is None:
            iterator = buf.get_iter_at_mark(start_mark)
            buf.insert(iterator, '\n')
        elif lexer != PLUGIN_INTERNAL_NONE_LEXER_ID:
            tokens = pygments.lex(code, lexer)

            formatter = GTKFormatter(style=style, start_mark=start_mark)
            pygments.format(tokens, formatter, buf)

    @staticmethod
    def _find_multiline_matches(text):
        start = None
        matches = []
        # Less strict, allow prefixed whitespaces:
        # for i in re.finditer(r'(?:^|\n)[ |\t]*(```)\S*[ |\t]*(?:\n|$)',
        #     text, re.DOTALL):
        for i in re.finditer(r'(?:^|\n)(```)\S*(?:\n|$)', text, re.DOTALL):
            if start is None:
                start = i
            elif re.match(r'^\n```', i.group(0)) is not None:
                matches.append(
                    (start.start(), i.end(), text[start.start():i.end()]))
                start = None
            else:
                # not an end...
                continue
        return matches

    @staticmethod
    def _find_inline_matches(text):
        """
        Inline code is highlighted if the start marker is precedded by a start
        of line, a whitespace character or either of the other span markers
        defined in XEP-0393.
        The same applies mirrored to the end marker.
        """
        return [(i.start(1), i.end(1), i.group(1)) for i in
                re.finditer(r'(?:^|\s|\*|~|_)(`((?!`).+?)`)(?:\s|\*|~|_|$)',
                            text)]

    @staticmethod
    def _merge_match_groups(real_text, inline_matches, multiline_matches):
        it_inline = iter(inline_matches)
        it_multi = iter(multiline_matches)
        length = len(real_text)

        # Just to get cleaner code below...
        def _get_next(iterator):
            return next(iterator, (length, length, ''))

        # In order to simplify the process, we use the 'length' here.
        cur_inline = _get_next(it_inline)
        cur_multi = _get_next(it_multi)

        pos = 0

        # This will contain tuples with parts of the input and its
        # classification
        parts = []
        while pos < length:
            log.debug('-> in: %s', str(cur_inline))
            log.debug('-> mu: %s', str(cur_multi))

            # selected = (start, end, type)
            if cur_inline[0] < cur_multi[0]:
                selected = (cur_inline[0], cur_inline[1], MatchType.INLINE)
            elif cur_multi[0] < length:
                selected = (cur_multi[0], cur_multi[1], MatchType.MULTILINE)
            else:
                selected = (pos, length, MatchType.TEXT)
            log.debug('--> select: %s', str(selected))

            # Handle plain text string parts (and unforseen errors...)
            if pos < selected[0]:
                end = selected[0] if selected[0] != pos else selected[1]
                parts.append((real_text[pos:end], MatchType.TEXT))
                pos = selected[0]
            elif pos > selected[0]:
                log.error('Should not happen, position > found match.')

            # Cut out and append selected text segment
            parts.append((real_text[selected[0]:selected[1]], selected[2]))
            pos = selected[1]

            # Depending on the match type, we have to forward the iterators.
            # Also, forward the other one, if regions overlap or we took over.
            if selected[2] == MatchType.INLINE:
                if cur_multi[0] < cur_inline[1]:
                    cur_multi = _get_next(it_multi)
                cur_inline = _get_next(it_inline)
            elif selected[2] == MatchType.MULTILINE:
                if cur_inline[0] < cur_multi[1]:
                    cur_inline = _get_next(it_inline)
                cur_multi = _get_next(it_multi)

        return parts

    def process_text(self, real_text, other_tags, _graphics, iter_,
                     _additional):
        def _fix_newline(char, marker_len_no_newline, force=False):
            fixed = (marker_len_no_newline, '')
            if char == '\n':
                fixed = (marker_len_no_newline + 1, '')
            elif force:
                fixed = (marker_len_no_newline + 1, '\n')
            return fixed

        buf = self.textview.tv.get_buffer()

        # First, try to find inline or multiline code snippets
        inline_matches = self._find_inline_matches(real_text)
        multiline_matches = self._find_multiline_matches(real_text)

        if not inline_matches and not multiline_matches:
            log.debug('Stopping early, since there is no code block in it...')
            return

        iterator = iter_ if iter_ is not None else buf.get_end_iter()

        # Create a start marker with left gravity before inserting text.
        start_mark = buf.create_mark('SHP_start', iterator, True)
        end_mark = buf.create_mark('SHP_end', iterator, False)

        insert_newline_for_multiline = self._check_line_break(True)
        insert_newline_for_inline = self._check_line_break(False)

        split_text = self._merge_match_groups(
            real_text, inline_matches, multiline_matches)

        buf.begin_user_action()

        for num, (text_to_insert, match_type) in enumerate(split_text):
            language = None
            end_of_message = num == (len(split_text) - 1)

            if match_type == MatchType.TEXT:
                self.textview.detect_and_print_special_text(
                    text_to_insert, other_tags, graphics=_graphics,
                    iter_=iterator, additional_data=_additional)
            else:
                if match_type == MatchType.MULTILINE:
                    language_match = re.search(
                        '\n*```([^\n]*)\n', text_to_insert, re.DOTALL)

                    language = None if language_match is None \
                        else language_match.group(1)

                    language_len = 0 if language is None else len(language)

                    # We account the language word width for the front marker
                    front = _fix_newline(
                        text_to_insert[0],
                        3 + language_len,
                        insert_newline_for_multiline)
                    back = _fix_newline(
                        text_to_insert[-1],
                        3,
                        insert_newline_for_multiline and not end_of_message)
                else:
                    front = _fix_newline(
                        text_to_insert[0],
                        1,
                        insert_newline_for_inline)
                    back = _fix_newline(
                        text_to_insert[-1],
                        1,
                        insert_newline_for_inline and not end_of_message)

                marker_widths = (front[0], back[0])
                text_to_insert = ''.join([front[1], text_to_insert, back[1]])

                # Insertion invalidates iterator, let's use our start mark...
                self._insert_and_format_code(
                    buf, text_to_insert, language, marker_widths, start_mark,
                    end_mark, other_tags)

            iterator = buf.get_iter_at_mark(end_mark)
            # The current end of the buffer's contents is the start for the
            # next iteration
            buf.move_mark(start_mark, iterator)

        buf.delete_mark(start_mark)
        buf.delete_mark(end_mark)

        buf.end_user_action()

        # We have to make sure this is the last thing we do (i.e. no calls to
        # the other textview methods no more from here on), because the
        # print_special_text method is resetting the plugin_modified variable.
        self.textview.plugin_modified = True

    def _insert_and_format_code(self, buf, insert_text, language, marker,
                                start_mark, end_mark, other_tags=None):

        start_iter = buf.get_iter_at_mark(start_mark)

        if other_tags:
            buf.insert_with_tags_by_name(start_iter, insert_text, *other_tags)
        else:
            buf.insert(start_iter, insert_text)

        tag_start = buf.get_iter_at_mark(start_mark)
        tag_end = buf.get_iter_at_mark(end_mark)
        s_code = tag_start.copy()
        e_code = tag_end.copy()
        s_code.forward_chars(marker[0])
        e_code.backward_chars(marker[1])

        log.debug('full text between tags: %s.', tag_start.get_text(tag_end))

        self._format_code(buf, tag_start, s_code, tag_end, e_code, language)

        self.textview.plugin_modified = True

        # Set general code block format
        tag = Gtk.TextTag.new()
        bg_color = self._plugin_config['bgcolor']
        if self._plugin_config['bgcolor_override']:
            tag.set_property('background', bg_color)
            tag.set_property('paragraph-background', bg_color)
        tag.set_property('font', self._plugin_config['font'])
        buf.get_tag_table().add(tag)
        buf.apply_tag(tag, tag_start, tag_end)
